# -*- coding: utf-8 -*-

from parser.metric import Metric
from parser.utils import CEOR
import numpy as np
import torch
import torch.nn as nn
from torch import log, exp
import torch.nn.functional as F
from scipy.stats import entropy as calc_entropy
from math import log, sqrt

class Model(object):

    def __init__(self, vocab, parser):
        super(Model, self).__init__()

        self.vocab = vocab
        self.parser = parser
        self.criterion = nn.CrossEntropyLoss()

    def train(self, loader):
        self.parser.train()

        for words, tags, chars, feats, arcs, rels in loader:
            self.optimizer.zero_grad()

            mask = words.ne(self.vocab.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            s_arc, s_rel = self.parser(words, chars, tags)
            s_arc, s_rel = s_arc[mask], s_rel[mask]
            gold_arcs, gold_rels = arcs[mask], rels[mask]

            loss = self.get_loss(s_arc, s_rel, gold_arcs, gold_rels)
            loss.backward()
            nn.utils.clip_grad_norm_(self.parser.parameters(), 5.0)
            self.optimizer.step()
            self.scheduler.step()

    def get_loss(self, s_arc, s_rel, gold_arcs, gold_rels):
        s_rel = s_rel[torch.arange(len(s_rel)), gold_arcs]
        arc_loss = self.criterion(s_arc, gold_arcs)
        rel_loss = self.criterion(s_rel, gold_rels)
        loss = arc_loss + rel_loss
        return loss

    
    @torch.no_grad()
    def evaluate(self, loader, punct=True):
        self.parser.eval()

        loss, metric = 0, Metric()

        for words, tags, chars, feats, arcs, rels in loader:
            mask = words.ne(self.vocab.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            # ignore all punctuation if not specified
            if not punct:
                puncts = words.new_tensor(self.vocab.puncts)
                mask &= words.unsqueeze(-1).ne(puncts).all(-1)
            s_arc, s_rel = self.parser(words, chars, tags)
            s_arc, s_rel = s_arc[mask], s_rel[mask]
            gold_arcs, gold_rels = arcs[mask], rels[mask]
            ###################################################################
            #### pass lens to decode if forcing well-formed tree during train #
            ### lens = mask.sum(dim=1).tolist()                               #
            ### pred_arcs, pred_rels = self.decode(s_arc, s_rel, lens)        #
            ###################################################################
            pred_arcs, pred_rels = self.decode(s_arc, s_rel)
            if pred_arcs is None:
                continue
            loss += self.get_loss(s_arc, s_rel, gold_arcs, gold_rels)
            metric(pred_arcs, pred_rels, gold_arcs, gold_rels)
        loss /= len(loader)

        return loss, metric

    @torch.no_grad()
    def predict(self, loader):
        self.parser.eval()

        all_arcs, all_rels = [], []
        for words, tags, chars, feats in loader:
            mask = words.ne(self.vocab.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            lens = mask.sum(dim=1).tolist()
            s_arc, s_rel = self.parser(words, chars, tags)
            s_arc, s_rel = s_arc[mask], s_rel[mask]
            pred_arcs, pred_rels = self.decode(s_arc, s_rel, lens)

            all_arcs.extend(torch.split(pred_arcs, lens))
            all_rels.extend(torch.split(pred_rels, lens))
        all_arcs = [seq.tolist() for seq in all_arcs]
        all_rels = [self.vocab.id2rel(seq) for seq in all_rels]

        return all_arcs, all_rels

    def decode(self, s_arc, s_rel, lens=False):
      
        if lens:
            pred_arcs_np = np.array([]).astype(int)
            s_arc_soft = F.softmax(s_arc,dim=-1)
            length_count = 0
            for i, l in enumerate(lens):
                topL = l + length_count
                arcs = s_arc[length_count:topL,0:l+1]
                soft_arcs = F.softmax(arcs,dim=-1).cpu().numpy()
                pred_arcs_np= np.append(pred_arcs_np,CEOR(soft_arcs).astype(int))
                length_count += l
            pred_arcs = torch.from_numpy(pred_arcs_np).to(s_arc.device)
        else:
            if len(s_arc) > 0:
                pred_arcs =  s_arc.argmax(dim=-1)
            else:
                return None, None
        pred_rels = s_rel[torch.arange(len(s_rel)), pred_arcs].argmax(dim=-1)

        return pred_arcs, pred_rels
    
