# -*- coding: utf-8 -*-

import os
from datetime import datetime, timedelta, date
from parser import BiaffineParser, Model
from parser.metric import Metric
from parser.utils import Corpus, Embedding, Vocab
from parser.utils.data import TextDataset, batchify
import torch
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR
import random

class Flip(object):

    def add_subparser(self, name, parser):
        subparser = parser.add_parser(
            name, help='Train a model.'
        )
        subparser.add_argument('--buckets', default=64, type=int,
                               help='max num of buckets to use')
        subparser.add_argument('--punct', action='store_true',
                               help='whether to include punctuation')
        subparser.add_argument('--ftrain', default='NONE',
                               help='path to train file')
        subparser.add_argument('--fsilver', default='NONE',
                               help='path to silver train file')
        subparser.add_argument('--fdev', default='NONE',
                               help='path to dev file')
        subparser.add_argument('--ftest', default='NONE',
                               help='path to test file')
        subparser.add_argument('--fembed', default=False,
                               help='path to pretrained embeddings')
        subparser.add_argument('--n_char_embed', default=100, type=int,
                               help='char embedding size')
        subparser.add_argument('--unk', default='unk',
                               help='unk token in pretrained embeddings')
        subparser.add_argument('--pflip', default=0.5, type=float,
                               help='probability of flipping to silver data')
        subparser.add_argument('--withhold', action='store_true', default=False, help='Whether to withhold 10% of gold data or not.')
        return subparser

    def __call__(self, config):
        print("Preprocess the data")
        train = Corpus.load(config.ftrain)
        if config.withhold:
            numberToWithold = int(len(train.sentences)/50)
            if numberToWithold < 5:
                numberToWithold = 5
            print("Withholding " + str(numberToWithold) + " gold training sentences.")
            train.sentences = train.sentences[:-numberToWithold]
        silver = Corpus.load(config.fsilver)
        all_training= Corpus(train.sentences+silver.sentences)
        
        dev = Corpus.load(config.fdev)
        test = Corpus.load(config.ftest)
        
        if os.path.exists(config.vocab):
            vocab = torch.load(config.vocab)
        else:
            vocab = Vocab.from_corpus(corpus=all_training, min_freq=10)
            if config.fembed:
                vocab.read_embeddings(Embedding.load(config.fembed, config.unk))
            else:
                vocab.randomly_initialise_embeddings(config.n_embed)
            torch.save(vocab, config.vocab)
        config.update({
            'n_words': vocab.n_train_words,
            'n_tags': vocab.n_tags,
            'n_chars': vocab.n_chars,
            'n_rels': vocab.n_rels,
            'pad_index': vocab.pad_index,
            'unk_index': vocab.unk_index
        })
        print(vocab)
        


        print("Load the dataset")
        trainset = TextDataset(vocab.numericalize(train))
        silverset = TextDataset(vocab.numericalize(silver))
        devset = TextDataset(vocab.numericalize(dev))
        testset = TextDataset(vocab.numericalize(test))
        # set the data loaders
        train_loader = batchify(dataset=trainset,
                                batch_size=config.batch_size,
                                n_buckets=config.buckets,
                                shuffle=True)
        silver_train_loader = batchify(dataset=silverset,
                                batch_size=config.batch_size,
                                n_buckets=config.buckets,
                                shuffle=True)
        dev_loader = batchify(dataset=devset,
                              batch_size=config.batch_size,
                              n_buckets=config.buckets)
        test_loader = batchify(dataset=testset,
                               batch_size=config.batch_size,
                               n_buckets=config.buckets)
        print(f"{'train:':6} {len(trainset):5} sentences in total, "
              f"{len(train_loader):3} batches provided")
        print(f"{'silver:':6} {len(silverset):5} sentences in total, "
              f"{len(silver_train_loader):3} batches provided")
        print(f"{'dev:':6} {len(devset):5} sentences in total, "
              f"{len(dev_loader):3} batches provided")
        print(f"{'test:':6} {len(testset):5} sentences in total, "
              f"{len(test_loader):3} batches provided")

        print("Create the model")
        parser = BiaffineParser(config, vocab.embeddings)
        if torch.cuda.is_available():
            parser = parser.cuda()
        print(f"{parser}\n")

        model = Model(vocab, parser)

        total_time = timedelta()
        best_e, best_metric = 1, Metric()
        
        model.optimizer = Adam(model.parser.parameters(),
                               config.lr,
                               (config.beta_1, config.beta_2),
                               config.epsilon)
        model.scheduler = ExponentialLR(model.optimizer,
                                        config.decay ** (1 / config.steps))
        logname=config.model.split(".")[0] + ".log"
        with open(logname, "w") as log:
            log.write(str(date.today()))
            log.write("\nepoch\ttrain\tdev\tsilver\tdata-type\ttime\n")
        for epoch in range(1, config.epochs + 1):
            start = datetime.now()
            # train one epoch and update the parameters
            if config.pflip > random.random():
                print("Using gold data.")
                train_data_type = "gold"
                model.train(train_loader)
            else:
                print("Using silver data.")
                train_data_type = "silver"
                model.train(silver_train_loader)
            print(f"Epoch {epoch} / {config.epochs}:")
            loss, train_metric = model.evaluate(train_loader, config.punct)
            print(f"{'train:':6} Loss: {loss:.4f} {train_metric}")
            loss, dev_metric = model.evaluate(dev_loader, config.punct)
            print(f"{'dev:':6} Loss: {loss:.4f} {dev_metric}")
            loss, silver_metric = model.evaluate(silver_train_loader, config.punct)
            print(f"{'silver:':6} Loss: {loss:.4f} {silver_metric}")
            t = datetime.now() - start
            with open(logname, "a") as log:
                log_output = [str(epoch), str(train_metric.score), str(dev_metric.score),
                              str(silver_metric.score), train_data_type, str(t.total_seconds())]
                log.write("\t".join(log_output))
                log.write("\n")

            # save the model if it is the best so far
            if dev_metric > best_metric:
                best_e, best_metric = epoch, dev_metric
                model.parser.save(config.model)
                print(f"{t}s elapsed (saved)\n")
            else:
                print(f"{t}s elapsed\n")
            total_time += t
        model.parser = BiaffineParser.load(config.model)
        loss, metric = model.evaluate(test_loader, config.punct)
        
        print(f"max score of dev is {best_metric.score:.2%} at epoch {best_e}")
        print(f"the score of test at epoch {best_e} is {metric.score:.2%}")
        print(f"average time of each epoch is {total_time / epoch}s")
        print(f"{total_time}s elapsed")
