# -*- coding: utf-8 -*-

import os, math
from datetime import datetime, timedelta, date
from parser import BiaffineParser, Model
from parser import LanguageModel
from parser.metric import Metric
from parser.utils import Corpus, Embedding, Vocab
from parser.utils.data import TextDataset, batchify
import torch
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR
from parser.utils.seq_lab import seq_lab_decoder_corpus
from parser.utils.calc_as import calc_attachment_scores

class Pretrain(object):

    def add_subparser(self, name, parser):
        subparser = parser.add_parser(
            name, help='Train a model.'
        )
        subparser.add_argument('--buckets', default=64, type=int,
                               help='max num of buckets to use')
        subparser.add_argument('--punct', action='store_true',
                               default=True,
                               help='whether to include punctuation')
        subparser.add_argument('--ftrain', default='data/ptb/train.conllx',
                               help='path to train file')
        subparser.add_argument('--fdev', default='data/ptb/dev.conllx',
                               help='path to dev file')
        subparser.add_argument('--fembed', default=False,
                               help='path to pretrained embeddings')
        subparser.add_argument('--n_char_embed', default=100, type=int,
                               help='char embedding size')
        subparser.add_argument('--unk', default='unk',
                               help='unk token in pretrained embeddings')
        subparser.add_argument('--use_pretraining', default=False,
                               help='whether or not to use pretrained parameters')
        subparser.add_argument('--withhold', action='store_true', default=False, help='Whether to withhold 10% of gold data or not.')
        subparser.add_argument('--fsilver', default=None,
                               help='path to silver train file')
        subparser.add_argument('--loss_weight_bi', default=0.1, type=float,
                               help='weight of biaffine loss')
        subparser.add_argument('--loss_weight_sl', default=1.0, type=float,
                               help='weight of sequence labelling loss')
        subparser.add_argument('--load_vocab', default=False, action="store_true",
                               help='weight of biaffine loss')
        subparser.add_argument("--patience", default=15, type=int,
                                help="number of epochs to wait without seeing progess for early stopping")
        
        return subparser

    def __call__(self, config):
        print("Preprocess the data")
        x = 0
        while x != 10000000:
            x+=1
        if math.isnan(config.loss_weight_sl):
            config.loss_weight_sl = 1 - config.loss_weight_bi
        train = Corpus.load(config.ftrain)
        dev = Corpus.load(config.fdev)
        if config.withhold:
            numberToWithold = int(len(train.sentences)/50)
            if numberToWithold < 5:
                numberToWithold = 5
            print("Withholding " + str(numberToWithold) + " gold training sentences.")
            train.sentences = train.sentences[:-numberToWithold]
        if config.fsilver:
            silver = Corpus.load(config.fsilver)
            train.sentences += silver.sentences

        
        if config.load_vocab and os.path.exists(config.vocab):
            vocab = torch.load(config.vocab)
        else:
            path = config.vocab + '.pretraining' if config.use_pretraining else None
            vocab = Vocab.from_corpus(corpus=train, min_freq=10, path=path)
            if config.fembed:
                vocab.read_embeddings(Embedding.load(config.fembed, config.unk, config.n_embed))
            else:
                vocab.randomly_initialise_embeddings(config.n_embed)
            torch.save(vocab, config.vocab)
            
        config.update({
            'n_words': vocab.n_train_words,
            'n_tags': vocab.n_tags,
            'n_chars': vocab.n_chars,
            'n_rels': vocab.n_rels,
            'n_enc': vocab.n_enc,
            'pad_index': vocab.pad_index,
            'unk_index': vocab.unk_index
        })
        print(vocab)
        exit()
        print("Load the dataset")
        trainset = TextDataset(vocab.numericalize(train))
        devset = TextDataset(vocab.numericalize(dev))
        # set the data loaders
        train_loader = batchify(dataset=trainset,
                                batch_size=config.batch_size,
                                n_buckets=config.buckets,
                                shuffle=True)
        dev_loader = batchify(dataset=devset,
                              batch_size=config.batch_size,
                              n_buckets=config.buckets)
        print(f"{'train:':6} {len(trainset):5} sentences in total, "
              f"{len(train_loader):3} batches provided")
        print(f"{'dev:':6} {len(devset):5} sentences in total, "
              f"{len(dev_loader):3} batches provided")
        
        print("Create the model")
        parser = BiaffineParser(config, vocab.embeddings)

        if config.use_pretraining:
            print("Set pretrained weights")
            path = config.model + '.pretraining'
            lm_state_dict = LanguageModel.load(path).state_dict()
            for item in lm_state_dict.keys():
                if item in parser.state_dict().keys():
                    parser.state_dict()[item] = lm_state_dict[item]

        if torch.cuda.is_available():
            parser = parser.cuda()
        print(f"{parser}\n")
        
        model = Model(vocab, parser)
        total_time = timedelta()
        best_epoch, best_metric, best_enc_f1 = 1, Metric(), -1
        best_seq_las = -1.
        model.optimizer = Adam(model.parser.parameters(),
                               config.lr,
                               (config.beta_1, config.beta_2),
                               config.epsilon)
        model.scheduler = ExponentialLR(model.optimizer,
                                        config.decay ** (1 / config.steps))
        log_path = "performance_logs/" + config.model + ".log"
        if config.loss_weight_sl != 0:
            with open(log_path, "w") as log:
                #log.write(str(date.today()))
                log.write("###epoch\ttrain-enc\ttrain-lab\tdev-enc\tdev-lab\ttime\n")
        for epoch in range(1, config.epochs + 1):
            start = datetime.now()
            # train one epoch and update the parameters
            model.train(train_loader)

            print(f"Epoch {epoch} / {config.epochs}:")
            loss, train_metric, train_enc_f1, train_lab_f1 = model.evaluate(train_loader, config.punct)
            print(f"{'train:':6}  Loss: {loss:.4f} {train_metric} \n \tBracket ACC: {train_enc_f1:.2f}%  Label ACC: {train_lab_f1:.2f}%" )
            loss, dev_metric, dev_enc_f1, dev_lab_f1 = model.evaluate(dev_loader, config.punct)
            print(f"{'dev:':6}  Loss: {loss:.4f} {dev_metric} \n \tBracket ACC: {dev_enc_f1:.2f}%  Label ACC: {dev_lab_f1:.2f}%" )
            if config.loss_weight_sl == 0:
                dev_seq_las = dev_metric.las*100
            else:
                pred_encs_dev, pred_rels_dev = model.predict_train_time(dev_loader)
                pred_dev_heads, pred_dev_rels = seq_lab_decoder_corpus(pred_encs_dev, pred_rels_dev)
                dev_seq_uas, dev_seq_las = calc_attachment_scores(dev.heads, dev.rels, pred_dev_heads, pred_dev_rels)
                print(f"{'dev seq lab:':6} UAS: {dev_seq_uas:.2f}%  LAS: {dev_seq_las:.2f}%")
            t = datetime.now() - start
            if config.loss_weight_sl != 0:
                with open(log_path, "a") as log:
                    log_output = [str(epoch),
                                  str(train_enc_f1),
                                  str(train_lab_f1),
                                  str(dev_enc_f1),
                                  str(dev_lab_f1),
                                  str(t.total_seconds())]
                    log.write("\t".join(log_output))
                    log.write("\n")
            if dev_seq_las > best_seq_las:
                print(f"***   {dev_seq_las:.2f} es mejor que {best_seq_las:.2f} entonces   ***")
                best_epoch, best_seq_las = epoch, dev_seq_las
                model.parser.save(config.model)
                print(f"{t}s elapsed (saved)\n")
            else:
                print(f"Current best: {best_seq_las:.2f} at epoch {best_epoch}")
                if epoch - best_epoch == config.patience:
                    print("*** Stopping early as no improvement after " + str(config.patience) + " epochs. ***")
                    break
                else:
                    print("T-"+str(config.patience - (epoch - best_epoch)) + " epochs until early stoppage.")
                print(f"{t}s elapsed\n")
            total_time += t
        model.parser = BiaffineParser.load(config.model)
                
        print(f"max score of dev is {best_seq_las:.2f} at epoch {best_epoch}")
        print(f"average time of each epoch is {total_time / epoch}s")
        print(f"{total_time}s elapsed")
