# -*- coding: utf-8 -*-

import os, math, csv
from datetime import datetime, timedelta, date
from parser import BiaffineParser, Model
from parser import LanguageModel
from parser.metric import Metric
from parser.utils import Corpus, Embedding, Vocab
from parser.utils.data import TextDataset, batchify
import torch
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR
from parser.utils.seq_lab import seq_lab_decoder_corpus
from parser.utils.calc_as import calc_attachment_scores

class Train(object):

    def add_subparser(self, name, parser):
        subparser = parser.add_parser(
            name, help='Train a model.'
        )
        subparser.add_argument('--punct', action='store_false',
                               default=True,
                               help='whether to include punctuation')
        subparser.add_argument('--ftrain', default='data/ptb/train.conllx',
                               help='path to train file')
        subparser.add_argument('--fdev', default='data/ptb/dev.conllx',
                               help='path to dev file')
        subparser.add_argument('--fembed', default=False,
                               help='path to pretrained embeddings')
        subparser.add_argument('--n_char_embed', default=100, type=int,
                               help='char embedding size')
        subparser.add_argument('--unk', default='unk',
                               help='unk token in pretrained embeddings')
        subparser.add_argument('--load_vocab', default=False, action="store_true",
                               help='weight of biaffine loss')
        #subparser.add_argument("--patience", default=15, type=int,
        #                        help="number of epochs to wait without seeing progess for early stopping")
        subparser.add_argument("--char_lemma", action="store_true", default=False)
        subparser.add_argument("--log_path", default=None, help="path where csv will be saved with training details")
        subparser.add_argument("--n_lstm_nodes", default=None, type=int)
        subparser.add_argument("--n_lstm_layers", default=None, type=int)
        subparser.add_argument("--tutti", action="store_true", default=False)
        return subparser

    def __call__(self, config):
        print("patience:", config.patience)
        config.n_lstm_hidden = config.n_lstm_nodes
        config.n_lstm_layers = config.n_lstm_layers
        print("Preprocess the data")
        train = Corpus.load(config.ftrain)
        dev = Corpus.load(config.fdev)
        if config.load_vocab and os.path.exists(config.vocab):
            vocab = torch.load(config.vocab)
        else:
            vocab = Vocab.from_corpus(corpus=train, min_freq=5, char_lemma=config.char_lemma)
            if config.fembed:
                embeddings_dict = Embedding.load(config.fembed, config.unk, config.n_embed)
                vocab.read_embeddings(embeddings_dict)
            else:
                vocab.randomly_initialise_embeddings(config.n_embed)
            torch.save(vocab, config.vocab)
        config.update({
            'n_words': vocab.n_words,
            'n_chars': vocab.n_chars,
            'n_rels': vocab.n_rels,
            'n_enc': vocab.n_enc,
            'pad_index': vocab.pad_index,
            'unk_index': vocab.unk_index
        })
        print(vocab)

        print("Load the dataset")
        trainset = TextDataset(vocab.numericalize(train,char_lemma=config.char_lemma))
        devset = TextDataset(vocab.numericalize(dev, char_lemma=config.char_lemma))
        # set the data loaders
        train_loader = batchify(dataset=trainset,
                                batch_size=config.batch_size,
                                n_buckets=config.buckets,
                                shuffle=True)
        dev_loader = batchify(dataset=devset,
                              batch_size=config.batch_size,
                              n_buckets=config.buckets)
        print(f"{'train:':6} {len(trainset):5} sentences in total, "
              f"{len(train_loader):3} batches provided")
        print(f"{'dev:':6} {len(devset):5} sentences in total, "
              f"{len(dev_loader):3} batches provided")
        
        print("Create the model")
        parser = BiaffineParser(config, vocab.embeddings)

        if torch.cuda.is_available():
            parser = parser.cuda()
        print(f"{parser}\n")
        
        model = Model(vocab, parser)
        total_time = timedelta()
        best_epoch, best_metric, best_enc_f1 = 1, Metric(), -1
        model.optimizer = Adam(model.parser.parameters(),
                               config.lr,
                               (config.beta_1, config.beta_2),
                               config.epsilon)
        model.scheduler = ExponentialLR(model.optimizer,
                                        config.decay ** (1 / config.steps))
        #log_path = "performance_logs/" + config.model + ".log"
        best_las = -99.9
        total_params = sum(p.numel() for p in model.parser.parameters())
        print ("Total number of parameters:", total_params,"\n")
        for epoch in range(1, config.epochs + 1):
            print(f"Epoch {epoch} / {config.epochs}:")
            start = datetime.now()
            # train one epoch and update the parameters
            model.train(train_loader)            
            loss,  train_enc_f1, train_lab_f1 = model.evaluate(train_loader, config.punct)
            print(f"{'train:':6}  Loss: {loss:.4f} \n \tBracket ACC: {train_enc_f1:.2f}%  Label ACC: {train_lab_f1:.2f}%" )
            loss,  dev_enc_f1, dev_lab_f1 = model.evaluate(dev_loader, config.punct)
            print(f"{'dev:':6}  Loss: {loss:.4f} \n \tBracket ACC: {dev_enc_f1:.2f}%  Label ACC: {dev_lab_f1:.2f}%" )
            pred_encs_dev, pred_rels_dev = model.predict_train_time(dev_loader)
            pred_dev_heads, pred_dev_rels = seq_lab_decoder_corpus(pred_encs_dev, pred_rels_dev)
            dev_uas, dev_las = calc_attachment_scores(dev.heads, dev.rels, pred_dev_heads, pred_dev_rels)
            print(f"{'     ':6}  UAS: {dev_uas:.2f}%  LAS: {dev_las:.2f}%")
            t = datetime.now() - start
            if dev_las > best_las:
                print(f"***   {dev_las:.2f} es mejor que {best_las:.2f} entonces   ***")
                best_epoch, best_uas, best_las = epoch, dev_uas, dev_las
                
                model.parser.save(config.model)
                print(f"{t}s elapsed (saved)\n")
            else:
                print(f"Current best: {best_las:.2f} at epoch {best_epoch}")
                if epoch - best_epoch == config.patience:
                    print("*** Stopping early as no improvement after " + str(config.patience) + " epochs. ***")
                    break
                else:
                    print("T-"+str(config.patience - (epoch - best_epoch)) + " epochs until early stoppage.")
                print(f"{t}s elapsed\n")
            total_time += t
        
                
        

        print(f"max score of dev is {best_las:.2f} at epoch {best_epoch}")
        print(f"average time of each epoch is {total_time / epoch}s")
        print(f"{total_time}s elapsed")

        
        if config.log_path is None:
            exit()
        if os.path.exists(config.log_path):
            parser = BiaffineParser.load(config.model)
            pred_encs_tr, pred_rels_tr = model.predict_train_time(train_loader)
            pred_tr_heads, pred_tr_rels = seq_lab_decoder_corpus(pred_encs_tr, pred_rels_tr)
            best_train_uas, best_train_las = calc_attachment_scores(train.heads, train.rels, pred_tr_heads, pred_tr_rels)
            with open(config.log_path, mode="a") as output:
                output_writer = csv.writer(output, delimiter=',')
                output_writer.writerow(["seqlab-"+config.model.split("/")[-1], #.split(".m")[0],
                                        str(len(trainset)),
                                        str(len(devset)),
                                        str(epoch),
                                        str(total_time.total_seconds()),
                                        str(total_params),
                                        str(vocab.n_words),
                                        str(vocab.n_chars),
                                        str(best_train_uas),
                                        str(best_train_las),
                                        str(best_uas),
                                        str(best_las)])
                                       
        else:
            with open(config.log_path, mode="w") as output:
                output_writer = csv.writer(output, delimiter=',')
                output_writer.writerow(["Model", "Train Trees",
                                        "Dev Trees", "Epochs",
                                        "Training Time", "Parameters",
                                        "Form Vocab", "Char Vocab",
                                        "Train UAS", "Train LAS",
                                        "Dev UAS", "Dev LAS"])
                output_writer.writerow(["seqlab-"+config.model.split("/")[-1], #.split(".m")[0],
                                        str(len(trainset)),
                                        str(len(devset)),
                                        str(epoch),
                                        str(total_time.total_seconds()),
                                        str(total_params),
                                        str(vocab.n_words),
                                        str(vocab.n_chars),
                                        str(best_train_uas),
                                        str(best_train_las),
                                        str(best_uas),
                                        str(best_las)])
