# -*- coding: utf-8 -*-

from parser.metric import Metric
from parser.utils import CEOR
import numpy as np
import torch
import torch.nn as nn
from torch import log, exp
import torch.nn.functional as F
from scipy.stats import entropy as calc_entropy
from math import log, sqrt

class Model(object):

    def __init__(self, vocab, parser):
        super(Model, self).__init__()

        self.vocab = vocab
        self.parser = parser
        self.criterion = nn.CrossEntropyLoss()

    def calc_label_accuracy(self, g_enc_all, p_enc_all):
        correct = 0.
        tokens = 0.
        for g_enc, p_enc in zip(g_enc_all, p_enc_all):
            for g, p in zip(g_enc, p_enc):
                if g.item() == p.item():
                    correct += 1.
                tokens += 1
        f1 = correct / tokens
        return f1*100

    
    def train(self, loader):
        self.parser.train()

        for words, chars, tags, rels in loader:
            self.optimizer.zero_grad()

            mask = words.ne(self.vocab.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            
            s_enc, s_rel_sl = self.parser(words, chars)
           
            s_enc, s_rel_sl =  s_enc[mask], s_rel_sl[mask]
            gold_rels, gold_enc = rels[mask], tags[mask]
        
            loss = self.get_loss(s_rel_sl, s_enc, gold_rels, gold_enc)

            loss.backward()
            nn.utils.clip_grad_norm_(self.parser.parameters(), 5.0)
            self.optimizer.step()
            self.scheduler.step()
            
            
    @torch.no_grad()
    def evaluate(self, loader, punct=True):
        self.parser.eval()

        loss, enc_f1, lab_f1 = 0, 0., 0.
        all_pred_enc, all_pred_rels = [], []
        all_gold_enc, all_gold_rels = [], []
        for words, chars, tags, rels in loader:
            mask = words.ne(self.vocab.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            s_enc, s_rel_sl = self.parser(words, chars)
            s_enc, s_rel_sl = s_enc[mask], s_rel_sl[mask]
            gold_rels, gold_enc = rels[mask], tags[mask]
            pred_enc, pred_rels_sl = self.decode(s_enc, s_rel_sl)
            all_gold_enc.append(gold_enc)
            all_pred_enc.append(pred_enc)
            all_gold_rels.append(gold_rels)
            all_pred_rels.append(pred_rels_sl)
            loss += self.get_loss(s_rel_sl, s_enc, gold_rels, gold_enc)
        enc_f1 = self.calc_label_accuracy(all_gold_enc, all_pred_enc)
        lab_f1 = self.calc_label_accuracy(all_gold_rels, all_pred_rels)
        loss /= len(loader)
        return loss, enc_f1, lab_f1

    @torch.no_grad()
    def predict(self, loader):
        self.parser.eval()
        all_enc, all_rels = [], []
       
        for words, chars in loader:
            mask = words.ne(self.vocab.pad_index)
            mask[:, 0] = 0
            lens = mask.sum(dim=1).tolist()
            s_enc, s_rel_sl = self.parser(words, chars)
            s_enc, s_rel_sl = s_enc[mask], s_rel_sl[mask]
            pred_enc, pred_rels = self.decode(s_enc, s_rel_sl)
            all_enc.extend(torch.split(pred_enc, lens))
            all_rels.extend(torch.split(pred_rels, lens))
        all_rels = [self.vocab.id2rel(seq) for seq in all_rels]
        all_enc = [self.vocab.id2enc(seq) for seq in all_enc]

        return all_enc, all_rels
    
       
    @torch.no_grad()
    def predict_train_time(self, loader):
        ### data loader different at train time (includes values to predicted)
        self.parser.eval()
        all_enc, all_rels = [], []
        for words, chars, tags, rels in loader:
            mask = words.ne(self.vocab.pad_index)
            # ignore the first token of each sentence
            mask[:, 0] = 0
            lens = mask.sum(dim=1).tolist()
            s_enc, s_rels_sl = self.parser(words, chars)
            s_enc, s_rels_sl = s_enc[mask], s_rels_sl[mask]

            pred_enc, pred_rels = self.decode(s_enc, s_rels_sl)

            all_rels.extend(torch.split(pred_rels, lens))
            all_enc.extend(torch.split(pred_enc, lens))
        all_rels = [self.vocab.id2rel(seq) for seq in all_rels]
        all_enc = [self.vocab.id2enc(seq) for seq in all_enc]
        return all_enc, all_rels
    
    
    def get_loss(self, s_rel_sl, s_enc, gold_rels, gold_enc):
        enc_loss = self.criterion(s_enc, gold_enc)
        rel_sl_loss = self.criterion(s_rel_sl, gold_rels)
        seq_lab_loss = enc_loss + rel_sl_loss
        loss = seq_lab_loss
        return loss
    
    
    def decode(self, s_enc, s_rel):
        pred_enc =  s_enc.argmax(dim=-1)
        pred_rel =  s_rel.argmax(dim=-1)
        return  pred_enc, pred_rel
    
