# -*- coding: utf-8 -*-

from collections import namedtuple
from operator import itemgetter
from parser.utils.seq_lab import seq_lab_encoder


Sentence = namedtuple(typename='Sentence',
                      field_names=['ID', 'FORM', 'LEMMA', 'CPOS',
                                   'POS', 'FEATS', 'HEAD', 'DEPREL',
                                   'PHEAD', 'PDEPREL', "ENC"],
                      defaults=[None]*11)

class RawCorpus(object):
    ROOT = '<ROOT>'
    PAD = '<PAD>'

    def __init__(self, sentences):
        super(RawCorpus, self).__init__()
        self.sentences = sentences

    def __len__(self):
        return len(self.sentences)


    def __getitem__(self, index):
        return self.sentences[index]

    @property
    def words(self):
        return [[self.ROOT] + list(sentence.FORM) for sentence in self]
    @property
    def lemmas(self):
        return [[self.ROOT] + list(sentence.LEMMA) for sentence in self]
    
    # for compatibility
    @property
    def tags(self):
        return [[self.ROOT] + list(sentence.CPOS) for sentence in self]

    @property
    def heads(self):
        return [[0] + list(map(int, sentence.HEAD)) for sentence in self]

    @property
    def rels(self):
        return [[self.ROOT] + list(sentence.DEPREL) for sentence in self]

    @classmethod
    def load(cls, fname):
        start, sentences = 0, []
        with open(fname, 'r') as f:
            lines = [line for line in f]
        for line in lines:
            line = line.strip().split()
            num_tokens = len(line)
            sentence = Sentence(None,
                                line,
                                [cls.PAD] * num_tokens,
                                [cls.PAD] * num_tokens,
                                [cls.PAD] * num_tokens,
                                [cls.PAD] * num_tokens,
                                [0] * num_tokens,
                                [cls.PAD] * num_tokens,
                                [cls.PAD] * num_tokens,
                                [cls.PAD] * num_tokens,
                                [cls.PAD] * num_tokens)                                
            sentences.append(sentence)
        corpus = cls(sentences)
        return corpus

    def save(self, fname):
        with open(fname, 'w') as f:
            f.write(f"{self}\n")


class Corpus(object):
    ROOT = '<ROOT>'

    def __init__(self, sentences):
        super(Corpus, self).__init__()
        self.entropies = False
        self.sentences = sentences

    def set_entropies(self, entropies):
        self.entropies = entropies
        
            
    def __len__(self):
        return len(self.sentences)

    def __repr__(self):
        return '\n'.join(
            '\n'.join('\t'.join(map(str, i))
                      for i in zip(*(f for f in sentence[:-1] if f))) + '\n'
            for sentence in self
        )

    def __getitem__(self, index):
        return self.sentences[index]

    @property
    def words(self):
        return [[self.ROOT] + list(sentence.FORM) for sentence in self]

    @property
    def lemmas(self):
        return [[self.ROOT] + list(sentence.LEMMA) for sentence in self]
    
    @property
    def tags(self):
        return [[self.ROOT] + list(sentence.CPOS) for sentence in self]

    @property
    def enc(self):
        return [[self.ROOT] + list(sentence.ENC) for sentence in self]

    @property
    def heads(self):
        return [[0] + list(map(int, sentence.HEAD)) for sentence in self]

    @property
    def rels(self):
        return [[self.ROOT] + list(sentence.DEPREL) for sentence in self]
    
    @heads.setter
    def heads(self, sequences):
        self.sentences = [sentence._replace(HEAD=sequence)
                          for sentence, sequence in zip(self, sequences)]

    @rels.setter
    def rels(self, sequences):
        self.sentences = [sentence._replace(DEPREL=sequence)
                          for sentence, sequence in zip(self, sequences)]


    @classmethod
    def load(cls, fname, raw=False):
        start, sentences = 0, []
        with open(fname, 'r') as f:
            lines = [line for line in f if not line.startswith("#")]
        for i, line in enumerate(lines):
            if len(line) <= 1:
                entries = [l.split("\t") for l in lines[start:i]]
                good_entries = []
                for entry in entries:
                    try:
                        int(entry[0])
                        entry[-1] = entry[-1].strip()
                        if raw:
                            entry.append("_")
                        good_entries.append(entry)    
                    except:
                        pass
                if not raw:
                    good_entries_with_enc = seq_lab_encoder(good_entries)
                    sentence = Sentence(*zip(*good_entries_with_enc))
                else:
                    sentence = Sentence(*zip(*good_entries))
                sentences.append(sentence)
                start = i + 1
       
        corpus = cls(sentences)

        return corpus

    def save(self, fname):
        with open(fname, 'w') as f:
            f.write(f"{self}\n")


                
