# -*- coding: utf-8 -*-

from collections import Counter

import regex
import torch


class Vocab(object):
    PAD = '<PAD>'
    UNK = '<UNK>'
    EOS = '<EOS>'

    def __init__(self, words, encs, chars, rels):
        self.pad_index = 0
        self.unk_index = 1

        self.words = [self.PAD, self.UNK] + sorted(words)
        self.encs = [self.PAD, self.UNK] + sorted(encs)
        self.chars = [self.PAD, self.UNK] + sorted(chars)
        self.rels = sorted(rels)

        self.word_dict = {word: i for i, word in enumerate(self.words)}
        self.enc_dict = {enc: i for i, enc in enumerate(self.encs)}
        self.char_dict = {char: i for i, char in enumerate(self.chars)}
        self.rel_dict = {rel: i for i, rel in enumerate(self.rels)}

        # ids of punctuation that appear in words
        self.puncts = sorted(i for word, i in self.word_dict.items()
                             if regex.match(r'\p{P}+$', word))

        self.n_words = len(self.words)
        self.n_chars = len(self.chars)
        self.n_rels = len(self.rels)
        self.n_train_words = self.n_words
        self.n_enc = len(self.encs)
        
    def __repr__(self):
        info = f"{self.__class__.__name__}: "
        info += f"{self.n_words} words, "
        info += f"{self.n_enc} encoding labels, "
        info += f"{self.n_chars} chars, "
        info += f"{self.n_rels} rels"

        return info

    def word2id(self, sequence):
        return torch.tensor([self.word_dict.get(word.lower(), self.unk_index)
                             for word in sequence])

    def enc2id(self, sequence):
        return torch.tensor([self.enc_dict.get(enc, self.unk_index)
                             for enc in sequence])
    
    def char2id(self, sequence, max_length=20):
        char_ids = torch.zeros(len(sequence), max_length, dtype=torch.long)
        for i, word in enumerate(sequence):
            ids = torch.tensor([self.char_dict.get(c, self.unk_index)
                                for c in word[:max_length]])
            char_ids[i, :len(ids)] = ids

        return char_ids

    def rel2id(self, sequence):
        return torch.tensor([self.rel_dict.get(rel, 0)
                             for rel in sequence])

    def id2rel(self, ids):
        return [self.rels[i] for i in ids]

    def id2enc(self, ids):
        return [self.encs[i] for i in ids]

                
    def read_embeddings(self, embed, smooth=True):
        # if the UNK token has existed in the pretrained,
        # then use it to replace the one in the vocab
        vocab_found = 0
        if embed.unk:
            self.UNK = embed.unk

        self.extend(embed.tokens)
        self.embeddings = torch.rand(self.n_words, embed.dim)
        for i, word in enumerate(self.words):
            if word in embed:
                self.embeddings[i] = embed[word]
                vocab_found += 1
                
        if smooth:
            self.embeddings /= torch.std(self.embeddings)
            
        print("Embedding coverage: " + str(round(100*vocab_found/float(self.n_words),3)))
        
            
    def randomly_initialise_embeddings(self, n_dim, smooth=True):
        self.embeddings = torch.randn(self.n_words, n_dim)
        if smooth:
            self.embeddings /= torch.std(self.embeddings)
            
    def extend(self, words):
        self.words += sorted(set(words).difference(self.word_dict))
        self.chars += sorted(set(''.join(words)).difference(self.char_dict))
        self.word_dict = {w: i for i, w in enumerate(self.words)}
        self.char_dict = {c: i for i, c in enumerate(self.chars)}
        self.puncts = sorted(i for word, i in self.word_dict.items()
                             if regex.match(r'\p{P}+$', word))
        self.n_words = len(self.words)
        self.n_chars = len(self.chars)

    def numericalize(self, corpus, training=True, char_lemma=False):
        words = [self.word2id(seq) for seq in corpus.words]
        if char_lemma:
            print("using lemma col for char seqs")
            chars = [self.char2id(seq) for seq in corpus.lemmas]
        else:
            chars = [self.char2id(seq) for seq in corpus.words]
        if not training:
            return words, chars
        encs = [self.enc2id(seq) for seq in corpus.enc]
        rels = [self.rel2id(seq) for seq in corpus.rels]

        return words, chars, encs, rels

    @classmethod
    def from_corpus(cls, corpus, min_freq=1, char_lemma=False):
        words = Counter(word.lower() for seq in corpus.words for word in seq)
        words = list(word for word, freq in words.items() if freq >= min_freq)
        if char_lemma:
            print("using lemma col for char emb")
            chars = list({char for seq in corpus.lemmas for char in ''.join(seq)})
        else:
            chars = list({char for seq in corpus.words for char in ''.join(seq)})
        encs = list({enc for seq in corpus.enc for enc in seq})
        rels = list({rel for seq in corpus.rels for rel in seq})
        vocab = cls(words, encs, chars, rels)
        return vocab
